/* Subscriber interface
Copyright (C) 2020 Jonathan Hanks

This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef PUB_SUB2_SUB_HH
#define PUB_SUB2_SUB_HH

#include <functional>
#include <memory>
#include <string>
#include <vector>

#include <cds-pubsub/common.hh>

namespace pub_sub
{

    namespace detail
    {
        struct SubscriberIntl;
    }

    /*!
     * @brief A SubMessage is the message returned from a producer to a
     * subscription handler.  It is a wrapper to hold subscription ids, message
     * keys, the message, and to automate cleanup of the message buffer.
     */
    class SubMessage
    {
    public:
        SubMessage( SubId id, KeyType key, Message msg )
            : sub_id_{ id }, key_{ key }, msg_{ std::move( msg ) }
        {
        }

        SubMessage( SubMessage&& other ) noexcept
            : sub_id_{ other.sub_id_ },
              key_{ other.key_ },
              msg_{ std::move( other.msg_ ) }
        {
            other.msg_.clear( );
        }

        SubMessage&
        operator=( SubMessage&& other ) noexcept
        {
            clear( );
            sub_id_ = other.sub_id_;
            key_ = other.key_;
            msg_ = other.msg_;
            other.msg_.clear( );
            return *this;
        }

        ~SubMessage( )
        {
            clear( );
        }

        SubId
        sub_id( ) const noexcept
        {
            return sub_id_;
        }

        KeyType
        key( ) const noexcept
        {
            return key_;
        }

        std::size_t
        size( ) const noexcept
        {
            return msg_.length;
        }

        void*
        data( ) const noexcept
        {
            return msg_.data( );
        }

        Message
        message( ) const noexcept
        {
            return msg_;
        }

    private:
        void
        clear( )
        {
        }
        SubId   sub_id_;
        KeyType key_;
        Message msg_;
    };

    /*!
     * @brief The signaturel/type of a subscription handler, a function that
     * receives SubMessage objects.
     */
    using SubHandler = std::function< void( SubMessage sub_message ) >;

    /*!
     * @brief The subscriber component of pub_sub.
     * @details A move only type that opens up and manages a connection to a
     * publisher.
     * It connects to the publisher and returns messages as they are received.
     * The lifecycle of the underlying transport (connection, reading, closing,
     * reconnecting, ...) is fully managed by the Subscriber.
     *
     * @note The publisher starts and controls a single background thread
     * which manages all connections and I/O.
     */
    class Subscriber
    {
    public:
        explicit Subscriber( );
        ~Subscriber( );

        Subscriber( const Subscriber& ) = delete;
        Subscriber( Subscriber&& ) = default;
        Subscriber& operator=( const Subscriber& ) = delete;
        Subscriber& operator=( Subscriber&& ) = default;

        /*!
         * @brief Install a set of debug hooks, call prior to subscribing
         */
        void SetupDebugHooks( SubDebugNotices* debug_hooks );

        /*!
         * @brief start a Subscription
         * @param address The address as a number (ie "127.0.0.1") to connect to
         * @param handler The callback for messages
         * @return The new SubId representing this subscription.
         */
        SubId subscribe( const std::string& address, SubHandler handler );

    private:
        std::unique_ptr< detail::SubscriberIntl > p_;
    };

} // namespace pub_sub

#endif // PUB_SUB2_SUB_HH
